/** @file
    File:       IccMpeXml.cpp

    Contains:   Implementation of MultiProcessElementType Element XML
                conversion functionality

    Version:    V1

    Copyright:   see ICC Software License
*/

/*
 * The ICC Software License, Version 0.2
 *
 *
 * Copyright (c) 2003-2010 The International Color Consortium. All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. In the absence of prior written permission, the names "ICC" and "The
 *    International Color Consortium" must not be used to imply that the
 *    ICC organization endorses or promotes products derived from this
 *    software.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE INTERNATIONAL COLOR CONSORTIUM OR
 * ITS CONTRIBUTING MEMBERS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the The International Color Consortium. 
 *
 *
 * Membership in the ICC is encouraged when this software is used for
 * commercial purposes. 
 *
 *  
 * For more information on The International Color Consortium, please
 * see <http://www.color.org/>.
 *  
 * 
 */
#include "IccMpeXml.h"
#include "IccUtilXml.h"

#include <cstring> /* C strings strcpy, memcpy ... */

#if defined(WIN32) || defined(WIN64)
#include <windows.h>
#ifdef GetClassName
#undef GetClassName
#endif
#endif

#ifdef USESAMPLEICCNAMESPACE
namespace sampleICC {
#endif

bool CIccMpeXmlUnknown::ToXml(std::string &xml, std::string blanks/* = ""*/)
{
  icUInt8Number *ptr = m_pData;

  char line[256];
  char buf[256];
  char fix[256];
  sprintf(line, "<UnknownElement Type=\"%s\" InputChannels=\"%d\" OutputChannels=\"%d\"", 
           icFixXml(fix, icGetSigStr(buf, GetType())), NumInputChannels(), NumOutputChannels());
  xml += blanks + line;

  if (m_nReserved) {
    sprintf(line, " Reserved=\"%u\"", m_nReserved);
    xml += buf;
  }
  xml += ">\n";

  icXmlDumpHexData(xml, blanks + "  ", m_pData, m_nSize);

  xml += blanks + "</UnknownElement>\n";
  return true;
}


bool CIccMpeXmlUnknown::ParseXml(xmlNode *pNode, std::string &parseStr)
{
  SetType((icElemTypeSignature)icXmlStrToSig(icXmlAttrValue(pNode, "type")));
  SetChannels(atoi(icXmlAttrValue(pNode, "InputChannels")), atoi(icXmlAttrValue(pNode, "OutputChannels")));

  if (pNode->children && pNode->children->type == XML_TEXT_NODE && pNode->children->content) {
    icUInt32Number nSize = icXmlGetHexDataSize((const char *)pNode->children->content);

    if (!SetDataSize(nSize, false)) 
      return false;

    icXmlGetHexData(m_pData, (const char *)pNode->children->content, nSize);
  }
  else {
    SetDataSize(0);
  }
  return true;
}


class CIccFormulaCurveSegmentXml : public CIccFormulaCurveSegment
{
public:
  CIccFormulaCurveSegmentXml(icFloatNumber start, icFloatNumber end) : CIccFormulaCurveSegment(start, end) {}

  bool ToXml(std::string &xml, std::string blanks/* = ""*/);
  bool ParseXml(xmlNode *pNode, std::string &parseStr);
};

static char* icSegPos(char *buf, icFloatNumber pos)
{
  if (pos==icMinFloat32Number)
    strcpy(buf, "-infinity");
  else if (pos==icMaxFloat32Number)
    strcpy(buf, "+infinity");
  else
    sprintf(buf, "%.8" ICFLOATSFX, pos);

  return buf;
}

icFloatNumber icGetSegPos(const char *str)
{
  if (!strncmp(str, "-inf", 4)) {
    return icMinFloat32Number;
  }
  if (!strncmp(str, "+inf", 4)) {
    return icMaxFloat32Number;
  }

  return (icFloatNumber)atof(str);
}


bool CIccFormulaCurveSegmentXml::ToXml(std::string &xml, std::string blanks)
{
  char buf[256];
  char line[256];

  sprintf(line, "<FormulaSegment Start=\"%s\"", icSegPos(buf, m_startPoint));
  xml += blanks + line;

  sprintf(line, " End=\"%s\"",icSegPos(buf, m_endPoint));
  xml += line;
    
  sprintf(line, " FunctionType=\"%d\"", m_nFunctionType);
  xml += line;

  if (m_nReserved) {
    sprintf(line, " Reserved=\"%d\"", m_nReserved);
    xml += line;
  }
  if (m_nReserved2) {
    sprintf(line, " Reserved2=\"%d\"", m_nReserved2);
    xml += line;
  }
  xml += ">\n";

  CIccFloatArray::DumpArray(xml, blanks+"  ", m_params, m_nParameters, icConvertFloat, 8);

  xml += blanks + "</FormulaSegment>\n";

  return true;
}


bool CIccFormulaCurveSegmentXml::ParseXml(xmlNode *pNode, std::string &parseStr)
{
  xmlAttr *funcType = icXmlFindAttr(pNode, "FunctionType");

  if (!funcType) {
    parseStr += "Bad FunctionType in Formula Segment\n";
    return false;
  }

  m_nReserved2 = atoi(icXmlAttrValue(pNode, "Reserved2"));
  m_nFunctionType = atoi(icXmlAttrValue(funcType));

  switch(m_nFunctionType) {
    case 0x0000:
      m_nParameters = 4;
      break;

    case 0x0001:
    case 0x0002:
      m_nParameters = 5;
      break;

    default:
      parseStr += "Unsupported FunctionType value in Formula Segment\n";
      return false;
  }

  CIccFloatArray args;

  if (!args.ParseArray(pNode->children))
    return false;

  if (args.GetSize()>m_nParameters)
    return false;

  if (m_params) {
    free(m_params);
  }

  if (m_nParameters) {
    m_params = (icFloatNumber*)malloc(m_nParameters * sizeof(icFloatNumber));
    if (!m_params)
      return false;

    memcpy(m_params, args.GetBuf(), m_nParameters * sizeof(icFloatNumber));
  }
  else
    m_params = NULL;

  return true;
}


class CIccSampledCurveSegmentXml : public CIccSampledCurveSegment
{
public:
  CIccSampledCurveSegmentXml(icFloatNumber start, icFloatNumber end) : CIccSampledCurveSegment(start, end) {}

  bool ToXml(std::string &xml, std::string blanks/* = ""*/);
  bool ParseXml(xmlNode *pNode, std::string &parseStr);
};

bool CIccSampledCurveSegmentXml::ToXml(std::string &xml, std::string blanks)
{
  char buf[256];
  char line[256];

  sprintf(line, "<SampledSegment Start=\"%s\"", icSegPos(buf, m_startPoint));
  xml += blanks + line;

  sprintf(line, " End=\"%s\">\n",icSegPos(buf, m_endPoint));
  xml += line;

  CIccFloatArray::DumpArray(xml, blanks+"  ", m_pSamples, m_nCount, icConvertFloat, 8);

  xml += blanks + "</SampledSegment>\n";

  return true;
}


bool CIccSampledCurveSegmentXml::ParseXml(xmlNode *pNode, std::string &parseStr)
{
  CIccFloatArray samples;

  if (!samples.ParseArray(pNode->children)) {
    return false;
  }

  if (!SetSize(samples.GetSize())) {
    return false;
  }

  memcpy(m_pSamples, samples.GetBuf(), samples.GetSize() * sizeof(icFloatNumber));

  return true;
}


class CIccSegmentedCurveXml : public CIccSegmentedCurve
{
public:
  bool ToXml(std::string &xml, std::string blanks/* = ""*/);
  bool ParseXml(xmlNode *pNode, std::string &parseStr);
};


bool CIccSegmentedCurveXml::ToXml(std::string &xml, std::string blanks)
{
  CIccCurveSegmentList::iterator seg;
  xml += blanks + "<SegmentedCurve>\n";
  for (seg = m_list->begin(); seg!=m_list->end(); seg++) {
    CIccCurveSegment* pSeg = *seg;
    if (pSeg && pSeg->GetType() == icSigFormulaCurveSeg) {
      if (!((CIccFormulaCurveSegmentXml*)pSeg)->ToXml(xml, blanks+"  "))
        return false;
    }
    else if (pSeg && pSeg->GetType() == icSigSampledCurveSeg) {
      if (!((CIccSampledCurveSegmentXml*)pSeg)->ToXml(xml, blanks+"  "))
        return false;
    }
    else
      return false;
  }
  xml += blanks + "</SegmentedCurve>\n";
  return true;
}


bool CIccSegmentedCurveXml::ParseXml(xmlNode *pNode, std::string &parseStr)
{
  m_list->clear();
  for (pNode=pNode->children; pNode; pNode=pNode->next) {
    if (pNode->type == XML_ELEMENT_NODE) {
      icFloatNumber start = icGetSegPos(icXmlAttrValue(pNode, "Start"));
      icFloatNumber end = icGetSegPos(icXmlAttrValue(pNode, "End"));

      if (!icXmlStrCmp(pNode->name, "FormulaSegment")) {
        CIccFormulaCurveSegmentXml *pSegXml = new CIccFormulaCurveSegmentXml(start, end);

        if (!pSegXml->ParseXml(pNode, parseStr)) {
          parseStr += "Unable to parse FormulaSegment\n";
          delete pSegXml;
          return false;
        }
        m_list->push_back(pSegXml);
      }
      else if (!icXmlStrCmp(pNode->name, "SampledSegment")) {
        CIccSampledCurveSegmentXml *pSegXml = new CIccSampledCurveSegmentXml(start, end);

        if (!pSegXml->ParseXml(pNode, parseStr)) {
          parseStr += "Unable to parse SampledSegment\n";
          delete pSegXml;
          return false;
        }
        m_list->push_back(pSegXml);
      }
      else 
        return false;
    }
  }
  return true;
}


bool CIccMpeXmlCurveSet::ToXml(std::string &xml, std::string blanks/* = ""*/)
{
  char line[256];
  sprintf(line, "<CurveSetElement InputChannels=\"%d\" OutputChannels=\"%d\"", NumInputChannels(), NumOutputChannels());
  xml += blanks + line;

  if (m_nReserved) {
    sprintf(line, " Reserved=\"%u\"", m_nReserved);
    xml += line;
  }
  xml += ">\n";

  int i;

  for (i=0; i<NumInputChannels(); i++) {
    CIccSegmentedCurveXml *ptr = (CIccSegmentedCurveXml*)m_curve[i];
    
    if (!(ptr->ToXml(xml, blanks+"  ")))
      return false;
  }

  xml += blanks + "</CurveSetElement>\n";
  return true;
}


bool CIccMpeXmlCurveSet::ParseXml(xmlNode *pNode, std::string &parseStr)
{
  icUInt16Number nChannels = (icUInt16Number)atoi(icXmlAttrValue(pNode, "InputChannels"));

  if (!nChannels || atoi(icXmlAttrValue(pNode, "OutputChannels")) != nChannels) {
    parseStr += "Invalid InputChannels or OutputChannels In CurveSetElement\n";
    return false;
  }

  SetSize(nChannels);
  int nIndex = 0;
  for (pNode = icXmlFindNode(pNode->children, "SegmentedCurve"), nIndex = 0;
       pNode && nIndex<nChannels;
       pNode=icXmlFindNode(pNode->next,"SegmentedCurve"), nIndex++) {
    CIccSegmentedCurveXml *pCurve = new CIccSegmentedCurveXml();

    if (!pCurve->ParseXml(pNode, parseStr)) {
      delete pCurve;
      return false;
    }
    if (!SetCurve(nIndex, pCurve))
      return false;
  }

  if (!pNode && nIndex == nChannels)
    return true;

  return false;
}


bool CIccMpeXmlMatrix::ToXml(std::string &xml, std::string blanks/* = ""*/)
{
  char buf[128];
  sprintf(buf, "<MatrixElement InputChannels=\"%d\" OutputChannels=\"%d\"", NumInputChannels(), NumOutputChannels());
  xml += blanks + buf;

  if (m_nReserved) {
    sprintf(buf, " Reserved=\"%u\"", m_nReserved);
    xml += buf;
  }
  xml += ">\n";

  int i, j, n;

  if (m_pMatrix) {
    xml += blanks + "  <MatrixData>\n";

    for (n=0, j=0; j<NumOutputChannels(); j++) {
      xml += blanks + "   ";
      for (i=0; i<NumInputChannels(); i++, n++) {
         sprintf(buf, " %.8f", m_pMatrix[n]);
         xml += buf;
      }
      xml += "\n";
    }
    xml += blanks + "  </MatrixData>\n";
  }

  if (m_pConstants) {
    xml += blanks + "  <ConstantData>\n";

    xml += blanks + "   ";
    for (i=0; i<NumOutputChannels(); i++) {
      sprintf(buf, " %.8f", m_pConstants[i]);
      xml += buf;
    }
    xml += "\n";

    xml += blanks + "  </ConstantData>\n";
  }

  xml += blanks + "</MatrixElement>\n";

  return true;
}


bool CIccMpeXmlMatrix::ParseXml(xmlNode *pNode, std::string &parseStr)
{
  icUInt16Number nInputChannels = atoi(icXmlAttrValue(pNode, "InputChannels"));
  icUInt16Number nOutputChannels = atoi(icXmlAttrValue(pNode, "OutputChannels"));
  if (!nInputChannels || !nOutputChannels) {
    parseStr += "Invalid InputChannels or OutputChannels In MatrixElement\n";
    return false;
  }

  xmlNode *pData;

  SetSize(nInputChannels, nOutputChannels);

  pData = icXmlFindNode(pNode->children, "MatrixData");
  if (pData) {
    if (!CIccFloatArray::ParseArray(m_pMatrix, m_nInputChannels*m_nOutputChannels, pData->children))
      return false;
  }
  pData = icXmlFindNode(pNode->children, "ConstantData");
  if (pData) {
    if (!CIccFloatArray::ParseArray(m_pConstants, m_nOutputChannels, pData->children))
      return false;
  }
  return true;
}


bool CIccMpeXmlCLUT::ToXml(std::string &xml, std::string blanks/* = ""*/)
{
  char attrs[256];

  if (m_nReserved) {
    sprintf(attrs, " InputChannels=\"%d\" OutputChannels=\"%d\" Reserved=\"%u\"", NumInputChannels(), NumOutputChannels(), m_nReserved);
  }
  else {
    sprintf(attrs, " InputChannels=\"%d\" OutputChannels=\"%d\"", NumInputChannels(), NumOutputChannels());
  }

  return icCLUTToXml(xml, m_pCLUT, icConvertFloat, blanks, true, attrs, "CLutElement");
}

extern CIccCLUT *icCLutFromXml(xmlNode *pNode, int nIn, int nOut, icConvertType nType, std::string &parseStr);

bool CIccMpeXmlCLUT::ParseXml(xmlNode *pNode, std::string &parseStr)
{
  m_nInputChannels = atoi(icXmlAttrValue(pNode, "InputChannels"));
  m_nOutputChannels = atoi(icXmlAttrValue(pNode, "OutputChannels"));

  if (!m_nInputChannels || !m_nOutputChannels) {
    parseStr += "Invalid InputChannels or OutputChannels In CLutElement\n";
    return false;
  }

  CIccCLUT *pCLut = icCLutFromXml(pNode, m_nInputChannels, m_nOutputChannels, icConvertFloat, parseStr);
  if (pCLut) {
    SetCLUT(pCLut);
    if (m_pCLUT)
      return true;
  }

  return false;
}


bool CIccMpeXmlBAcs::ToXml(std::string &xml, std::string blanks/* = ""*/)
{
  char line[256];
  char buf[256], fix[256];

  sprintf(line, "<BAcsElement InputChannels=\"%d\" OutputChannels=\"%d\" Signature=\"%s\"", NumInputChannels(), NumOutputChannels(),
                icFixXml(fix, icGetSigStr(buf, m_signature)));
  xml += blanks + line;

  if (m_nReserved) {
    sprintf(line, " Reserved=\"%u\"", m_nReserved);
    xml += buf;
  }

  if (m_pData && m_nDataSize) {
    icUInt8Number *ptr = m_pData;

    xml += ">\n";
  
    icXmlDumpHexData(xml, blanks+"  ", m_pData, m_nDataSize);

    xml += blanks + "</BAcsElement>\n";
  }
  else {
    xml += "/>\n";
  }

  return true;
}


bool CIccMpeXmlBAcs::ParseXml(xmlNode *pNode, std::string &parseStr)
{
  m_nInputChannels = atoi(icXmlAttrValue(pNode, "InputChannels"));
  m_nOutputChannels = atoi(icXmlAttrValue(pNode, "OutputChannels"));

  if (!m_nInputChannels || !m_nOutputChannels) {
    parseStr += "Invalid InputChannels or OutputChannels In BAcsElement\n";
    return false;
  }

  m_signature = icXmlStrToSig(icXmlAttrValue(pNode, "Signature"));

  if (pNode->children && pNode->children->type == XML_TEXT_NODE && pNode->children->content) {
    icUInt32Number nSize = icXmlGetHexDataSize((const char*)pNode->children->content);

    if (nSize) {
      if (!AllocData(nSize))
        return false;
      icXmlGetHexData(m_pData, (const char *)pNode->children->content, nSize);
    }
  }
  return true;
}


bool CIccMpeXmlEAcs::ToXml(std::string &xml, std::string blanks/* = ""*/)
{
  char line[256];
  char buf[256], fix[256];

  sprintf(line, "<EAcsElement InputChannels=\"%d\" OutputChannels=\"%d\" Signature=\"%s\"", NumInputChannels(), NumOutputChannels(),
    icFixXml(fix, icGetSigStr(buf, m_signature)));
  xml += blanks + line;

  if (m_nReserved) {
    sprintf(line, " Reserved=\"%u\"", m_nReserved);
    xml += buf;
  }

  if (m_pData && m_nDataSize) {
    icUInt8Number *ptr = m_pData;

    xml += ">\n";
    icXmlDumpHexData(xml, blanks+"  ", m_pData, m_nDataSize);
    xml += blanks + "</EAcsElement>\n";
  }
  else {
    xml += "/>\n";
  }

  return true;
}


bool CIccMpeXmlEAcs::ParseXml(xmlNode *pNode, std::string &parseStr)
{
  m_nInputChannels = atoi(icXmlAttrValue(pNode, "InputChannels"));
  m_nOutputChannels = atoi(icXmlAttrValue(pNode, "OutputChannels"));

  if (!m_nInputChannels || !m_nOutputChannels) {
    parseStr += "Invalid InputChannels or OutputChannels In EAcsElement\n";
    return false;
  }

  m_signature = icXmlStrToSig(icXmlAttrValue(pNode, "Signature"));

  if (pNode->children && pNode->children->type == XML_TEXT_NODE && pNode->children->content) {
    icUInt32Number nSize = icXmlGetHexDataSize((const char*)pNode->children->content);

    if (nSize) {
      if (!AllocData(nSize))
        return false;
      icXmlGetHexData(m_pData, (const char *)pNode->children->content, nSize);
    }
  }
  return true;
}


#ifdef USESAMPLEICCNAMESPACE
}
#endif
